""" Implementation of search interface to be used for tests where ElasticSearch is unavailable """

import copy
from collections import defaultdict
from datetime import datetime
import json
import os
import pytz

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

from search.elastic import RESERVED_CHARACTERS
from search.search_engine_base import SearchEngine
from search.utils import ValueRange, DateRange, _is_iterable


def json_date_to_datetime(json_date_string_value):
    ''' converts json date string to date object '''
    if "T" in json_date_string_value:
        if "." in json_date_string_value:
            format_string = "%Y-%m-%dT%H:%M:%S.%f"
        else:
            format_string = "%Y-%m-%dT%H:%M:%S"
        if json_date_string_value.endswith("Z"):
            format_string += "Z"

    else:
        format_string = "%Y-%m-%d"

    return datetime.strptime(
        json_date_string_value,
        format_string
    )


def _find_field(doc, field_name):
    """ find the dictionary field corresponding to the . limited name """
    if not isinstance(doc, dict):
        raise ValueError('Parameter `doc` should be a python dict object')

    if not isinstance(field_name, str):
        raise ValueError('Parameter `field_name` should be a string')

    immediate_field, remaining_path = field_name.split('.', 1) if '.' in field_name else (field_name, None)
    field_value = doc.get(immediate_field)

    if isinstance(field_value, dict):
        return _find_field(field_value, remaining_path)

    return field_value


def _filter_intersection(documents_to_search, dictionary_object, include_blanks=False):
    """
    Filters out documents that do not match all of the field values within the dictionary_object
    If include_blanks is True, then the document is considered a match if the field is not present
    """
    if not dictionary_object:
        return documents_to_search

    def value_matches(doc, field_name, field_value):
        """
        Does the document's field match the desired value, or is the field not present if include_blanks is True
        """
        compare_value = _find_field(doc, field_name)
        if compare_value is None:
            return include_blanks

        # if we have a string that we are trying to process as a date object
        if isinstance(field_value, (DateRange, datetime)):
            if isinstance(compare_value, str):
                compare_value = json_date_to_datetime(compare_value)

            field_has_tz_info = False
            if isinstance(field_value, DateRange):
                field_has_tz_info = (
                    (field_value.lower and field_value.lower.tzinfo is not None)
                    or (field_value.upper and field_value.upper.tzinfo is not None)
                )
            else:
                field_has_tz_info = field_value.tzinfo is not None

            if not field_has_tz_info:
                compare_value = compare_value.replace(tzinfo=None)
            elif compare_value.tzinfo is None:
                compare_value = compare_value.replace(tzinfo=pytz.UTC)

        if isinstance(field_value, ValueRange):
            return (
                (field_value.lower is None or compare_value >= field_value.lower)
                and (field_value.upper is None or compare_value <= field_value.upper)
            )
        if _is_iterable(compare_value) and not _is_iterable(field_value):
            return any(item == field_value for item in compare_value)

        if _is_iterable(field_value) and not _is_iterable(compare_value):
            return any(item == compare_value for item in field_value)

        if _is_iterable(compare_value) and _is_iterable(field_value):
            return any(str(item) in field_value for item in compare_value)

        return compare_value == field_value

    filtered_documents = documents_to_search
    for field_name, field_value in dictionary_object.items():
        filtered_documents = [d for d in filtered_documents if value_matches(d, field_name, field_value)]

    return filtered_documents


def _process_query_string(documents_to_search, query_string):
    """ keep the documents that contain at least one of the search strings provided """
    def _encode_string(string):
        """Encode a Unicode string in the same way as the Elasticsearch search engine."""
        string = string.translate(string.maketrans('', '', RESERVED_CHARACTERS))
        return string

    def has_string(dictionary_object, search_string):
        """ search for string in dictionary items, look down into nested dictionaries """
        for name in dictionary_object:
            if isinstance(dictionary_object[name], dict):
                return has_string(dictionary_object[name], search_string)
            if dictionary_object[name]:
                if search_string.lower() in _encode_string(dictionary_object[name].lower()):
                    return True
        return False

    search_strings = _encode_string(query_string).split(" ")
    documents_to_keep = []
    for search_string in search_strings:
        documents_to_keep.extend(
            [d for d in documents_to_search if "content" in d and has_string(d["content"], search_string)]
        )

    return documents_to_keep


def _process_exclude_dictionary(documents_to_search, exclude_dictionary):
    """ remove results that have fields that match in the exclude_dictionary """
    for exclude_property in exclude_dictionary:
        exclude_values = exclude_dictionary[exclude_property]
        if not isinstance(exclude_values, list):
            exclude_values = [exclude_values]
        documents_to_search = [
            document
            for document in documents_to_search
            if document.get(exclude_property) not in exclude_values
        ]
    return documents_to_search


def _count_aggregated_values(documents, aggregation_terms):
    """
    Calculate the counts for the aggregations provided:

    For each aggregation, count up the number of hits for each aggregation value, so
    that we can report back the breakdown of how many of each value there
    exist. Notice that the total is the total number of aggregation matches that
    we receive - a single document will get counted multiple times in the
    total if the aggregation field is multi-valued:

        e.g. a course may have a value for modes as ["honor", "validated"], and
        so the document will count as 1 towards the honor count, 1 towards the
        validated count, and 2 towards the total. (This may be a little
        surprising but matches the behaviour that elasticsearch presents)

    """
    aggregations = {}

    def process_aggregation(aggregate):
        """
        Find the values for this aggregation
        """
        aggregated_documents = [
            agg_document for agg_document in documents if aggregate in agg_document
        ]
        terms = defaultdict(int)

        def add_agg_value(agg_value):
            """
            Adds the discovered value to the counts for the selected aggregation.
            """
            if isinstance(agg_value, list):
                for individual_value in agg_value:
                    add_agg_value(individual_value)
            else:
                terms[agg_value] += 1

        for document in aggregated_documents:
            add_agg_value(document[aggregate])

        total = sum([terms[term] for term in terms])    # pylint: disable=consider-using-generator

        return total, terms

    for agg in aggregation_terms:
        total, terms = process_aggregation(agg)
        aggregations[agg] = {
            "total": total,
            "terms": terms,
        }

    return aggregations


class MockSearchEngine(SearchEngine):

    """
    Mock implementation of SearchEngine for test purposes
    """
    _mock_elastic = {}
    _disabled = False
    _file_name_override = None

    @classmethod
    def create_test_file(cls, file_name=None, index_content=None):
        """ creates test file from settings """
        if index_content:
            cls._mock_elastic = index_content
        else:
            cls._mock_elastic = {}
        if file_name:
            cls._file_name_override = file_name
        cls._write_to_file(create_if_missing=True)

    @classmethod
    def destroy_test_file(cls):
        """ creates test file from settings """
        file_name = cls._backing_file()
        if os.path.exists(file_name):
            os.remove(file_name)

        cls._file_name_override = None
        cls.destroy()
        cls.__disabled = False  # pylint: disable=unused-private-member

    @classmethod
    def _backing_file(cls, create_if_missing=False):
        """ return path to test file to use for backing purposes """
        backing_file_name = getattr(settings, "MOCK_SEARCH_BACKING_FILE", None)
        if cls._file_name_override:
            backing_file_name = cls._file_name_override

        if not backing_file_name:
            cls._disabled = False
            return None

        if create_if_missing or os.path.exists(backing_file_name):
            cls._disabled = False
            return backing_file_name

        cls._disabled = True
        return None

    @classmethod
    def _write_to_file(cls, create_if_missing=False):
        """ write the index dict to the backing file """
        file_name = cls._backing_file(create_if_missing)
        if file_name:
            with open(file_name, "w+", encoding='utf-8') as dict_file:
                json.dump(cls._mock_elastic, dict_file, cls=DjangoJSONEncoder)

    @classmethod
    def _load_from_file(cls):
        """ load the index dict from the contents of the backing file """
        file_name = cls._backing_file()
        if file_name and os.path.exists(file_name):
            with open(file_name, encoding='utf-8') as dict_file:
                cls._mock_elastic = json.load(dict_file)

    @staticmethod
    def _paginate_results(size, from_, raw_results):
        """ Give the correct page of results """
        results = raw_results
        if size:
            start = 0
            if from_ is not None:
                start = from_
            results = raw_results[start:start + size]

        return results

    @classmethod
    def load_index(cls, index_name):
        """ load the index, if necessary from the backed file """
        cls._load_from_file()
        if index_name not in cls._mock_elastic:
            cls._mock_elastic[index_name] = []
            cls._write_to_file()

        return cls._mock_elastic[index_name]

    @classmethod
    def add_documents(cls, index_name, sources):
        """
        Add documents to index.
        """
        cls.load_index(index_name).extend(sources)
        cls._write_to_file()

    @classmethod
    def remove_documents(cls, index_name, doc_ids):
        """
        Remove documents by id from index.
        """
        index = cls.load_index(index_name)
        cls._mock_elastic[index_name] = [d for d in index if "id" not in d or d["id"] not in doc_ids]
        cls._write_to_file()

    @classmethod
    def destroy(cls):
        """ Clean out the dictionary for test resets """
        cls._mock_elastic = {}
        cls._write_to_file()

    def __init__(self, index=None):
        super().__init__(index)
        MockSearchEngine.load_index(self.index_name)

    def index(self, sources, **kwargs):
        """
        Add/update documents to the index.
        """
        if not MockSearchEngine._disabled:
            doc_ids = [s["id"] for s in sources if "id" in s]
            MockSearchEngine.remove_documents(self.index_name, doc_ids)
            MockSearchEngine.add_documents(self.index_name, sources)

    def remove(self, doc_ids):  # pylint: disable=arguments-differ
        """
        Remove documents with given ids from the index.
        """
        if not MockSearchEngine._disabled:
            MockSearchEngine.remove_documents(self.index_name, doc_ids)

    def search(self,
               query_string=None,
               field_dictionary=None,
               filter_dictionary=None,
               exclude_dictionary=None,
               aggregation_terms=None,
               log_search_params=False,
               **kwargs):  # pylint: disable=too-many-arguments
        """
        Perform search upon documents within index.
        """
        if MockSearchEngine._disabled:
            return {
                "took": 10,
                "total": 0,
                "max_score": 0,
                "results": []
            }

        documents_to_search = []
        index = MockSearchEngine.load_index(self.index_name)
        documents_to_search.extend(index)

        if field_dictionary:
            documents_to_search = _filter_intersection(documents_to_search, field_dictionary)

        if filter_dictionary:
            documents_to_search = _filter_intersection(documents_to_search, filter_dictionary, True)

        if query_string:
            documents_to_search = _process_query_string(documents_to_search, query_string)

        # Support deprecated argument of exclude_ids
        if "exclude_ids" in kwargs:
            if not exclude_dictionary:
                exclude_dictionary = {}
            if "id" not in exclude_dictionary:
                exclude_dictionary["id"] = []
            exclude_dictionary["id"].extend(kwargs["exclude_ids"])

        if exclude_dictionary:
            documents_to_search = _process_exclude_dictionary(documents_to_search, exclude_dictionary)

        # Finally, find duplicates and give them a higher score
        def score_documents(documents_to_search):
            """ Apply scoring to documents that have multiple matches """
            search_results = []
            max_score = 0
            while documents_to_search:
                current_doc = documents_to_search[0]
                score = len([d for d in documents_to_search if d == current_doc])
                max_score = max(max_score, score)
                documents_to_search = [d for d in documents_to_search if d != current_doc]

                data = copy.copy(current_doc)
                search_results.append(
                    {
                        "score": score,
                        "data": data,
                    }
                )
            return search_results, max_score

        search_results, max_score = score_documents(documents_to_search)

        results = MockSearchEngine._paginate_results(
            kwargs["size"] if "size" in kwargs else None,
            kwargs["from_"] if "from_" in kwargs else None,
            sorted(search_results, key=lambda k: k["score"])
        )

        response = {
            "took": 10,
            "total": len(search_results),
            "max_score": max_score,
            "results": results
        }

        if aggregation_terms:
            response["aggs"] = _count_aggregated_values(documents_to_search, aggregation_terms)

        return response



#!/usr/bin/env python
# Some of the subclasses that get used as settings-overrides will yield this pylint
# error, but they do get used when included as part of the override_settings
""" Tests for search functionalty """

import copy
from datetime import datetime
import ddt

from django.core.cache import cache
from django.test import TestCase
from django.test.utils import override_settings
from elasticsearch import Elasticsearch

from search.api import course_discovery_search, NoSearchEngineError
from search.elastic import ElasticSearchEngine
from search.tests.utils import SearcherMixin, TEST_INDEX_NAME
from .mock_search_engine import MockSearchEngine


class DemoCourse:
    """ Class for dispensing demo courses """
    DEMO_COURSE_ID = "edX/DemoX/Demo_Course"
    DEMO_COURSE = {
        "start": datetime(2014, 2, 1),
        "number": "DemoX",
        "content": {
            "short_description": "Short description",
            "overview": "Long overview page",
            "display_name": "edX Demonstration Course",
            "number": "DemoX"
        },
        "course": "edX/DemoX/Demo_Course",
        "image_url": "/c4x/edX/DemoX/asset/images_course_image.jpg",
        "effort": "5:30",
        "id": DEMO_COURSE_ID,
        "enrollment_start": datetime(2014, 1, 1),
    }

    demo_course_count = 0

    @classmethod
    def get(cls, update_dict=None, remove_fields=None):
        """ get a new demo course """
        cls.demo_course_count += 1
        course_copy = copy.deepcopy(cls.DEMO_COURSE)
        if update_dict:
            if "content" in update_dict:
                course_copy["content"].update(update_dict["content"])
                del update_dict["content"]
            course_copy.update(update_dict)
        course_copy.update({"id": "{}_{}".format(course_copy["id"], cls.demo_course_count)})
        if remove_fields:
            for remove_field in remove_fields:
                if remove_field in course_copy:
                    del course_copy[remove_field]
        return course_copy

    @classmethod
    def reset_count(cls):
        """ go back to zero """
        cls.demo_course_count = 0

    @staticmethod
    def index(searcher, course_info):
        """ Adds course info dictionary to the index """
        searcher.index(sources=course_info)

    @classmethod
    def get_and_index(cls, searcher, update_dict=None, remove_fields=None):
        """ Adds course info dictionary to the index """
        cls.index(searcher, [cls.get(update_dict, remove_fields)])


@override_settings(SEARCH_ENGINE="search.tests.mock_search_engine.MockSearchEngine")
@override_settings(ELASTIC_FIELD_MAPPINGS={
    "start_date": {"type": "date"},
    "enrollment_start": {"type": "date"},
    "enrollment_end": {"type": "date"}
})
@override_settings(MOCK_SEARCH_BACKING_FILE=None)
@override_settings(COURSEWARE_CONTENT_INDEX_NAME=TEST_INDEX_NAME)
@override_settings(COURSEWARE_INFO_INDEX_NAME=TEST_INDEX_NAME)
# Any class that inherits from TestCase will cause too-many-public-methods pylint error
class TestMockCourseDiscoverySearch(TestCase, SearcherMixin):
    """
    Tests course discovery activities
    """

    @property
    def _is_elastic(self):
        """ check search engine implementation, to manage cleanup differently """
        return isinstance(self.searcher, ElasticSearchEngine)

    def setUp(self):
        super().setUp()
        # ignore unexpected-keyword-arg; ES python client documents that it can be used
        # pylint: disable=unexpected-keyword-arg
        if self._is_elastic:
            _elasticsearch = Elasticsearch()
            # Make sure that we are fresh
            _elasticsearch.indices.delete(index=TEST_INDEX_NAME, ignore=[400, 404])

            # remove cached property mappings (along with everything else)
            cache.clear()

            config_body = {}
            # ignore unexpected-keyword-arg; ES python client documents that it can be used
            _elasticsearch.indices.create(index=TEST_INDEX_NAME, ignore=400, body=config_body)
        else:
            MockSearchEngine.destroy()
        DemoCourse.reset_count()
        self._searcher = None

    def tearDown(self):
        # ignore unexpected-keyword-arg; ES python client documents that it can be used
        # pylint: disable=unexpected-keyword-arg
        if self._is_elastic:
            _elasticsearch = Elasticsearch()
            _elasticsearch.indices.delete(index=TEST_INDEX_NAME, ignore=[400, 404])
        else:
            MockSearchEngine.destroy()

        self._searcher = None
        super().tearDown()

    def test_course_list(self):
        """ No arguments to course_discovery_search should show all available courses"""
        results = course_discovery_search()
        self.assertEqual(results["total"], 0)

        DemoCourse.get_and_index(self.searcher)
        results = course_discovery_search()
        self.assertEqual(results["total"], 1)

    def test_course_matching(self):
        """ Make sure that matches within content can be located and processed """
        results = course_discovery_search("defensive")
        self.assertEqual(results["total"], 0)

        DemoCourse.get_and_index(self.searcher, {"content": {"short_description": "This is a defensive move"}})
        DemoCourse.get_and_index(self.searcher, {"content": {"overview": "Defensive teams often win"}})
        DemoCourse.get_and_index(self.searcher)

        results = course_discovery_search()
        self.assertEqual(results["total"], 3)
        results = course_discovery_search("defensive")
        self.assertEqual(results["total"], 2)

    def test_enroll_date(self):
        """
        Test that we don't show any courses that have no published enrollment date, or an enrollment date in the future
        """
        # demo_course_1 should be found cos it has a date that is valid
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": datetime(2014, 1, 1)})

        # demo_course_2 should not be found because it has enrollment_start date set explicitly to None
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": None})

        # demo_course_3 should not be found because it has enrollment_start date in the future
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": datetime(2114, 1, 1)})

        # demo_course_4 should not be found because it has no enrollment_start specification
        DemoCourse.get_and_index(self.searcher, {}, ["enrollment_start"])

        results = course_discovery_search()
        self.assertEqual(results["total"], 1)

        additional_course = DemoCourse.get()
        DemoCourse.index(self.searcher, [additional_course])

        results = course_discovery_search()
        self.assertEqual(results["total"], 2)

        # Mark the course as having ended enrollment
        additional_course["enrollment_end"] = datetime(2015, 1, 1)
        DemoCourse.index(self.searcher, [additional_course])

        results = course_discovery_search()
        self.assertEqual(results["total"], 1)

    @override_settings(SEARCH_SKIP_ENROLLMENT_START_DATE_FILTERING=True)
    def test_enroll_date_skipped(self):
        """
        Test feature flag to skip enrollment start date filtering for course search
        """
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": datetime(2014, 1, 1)})
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": None})
        DemoCourse.get_and_index(self.searcher, {"enrollment_start": datetime(2114, 1, 1)})
        DemoCourse.get_and_index(self.searcher, {}, ["enrollment_start"])
        results = course_discovery_search()
        self.assertEqual(results["total"], 4)

    def test_discovery_field_matching(self):
        """ Test that field specifications only show those results with the desired field values """
        DemoCourse.get_and_index(self.searcher, {"org": "OrgA"})
        DemoCourse.get_and_index(self.searcher, {"org": "OrgB"})

        results = course_discovery_search()
        self.assertEqual(results["total"], 2)

        results = course_discovery_search(field_dictionary={"org": "OrgA"})
        self.assertEqual(results["total"], 1)
        self.assertEqual(results["results"][0]["data"]["org"], "OrgA")

        results = course_discovery_search(field_dictionary={"org": "OrgB"})
        self.assertEqual(results["total"], 1)
        self.assertEqual(results["results"][0]["data"]["org"], "OrgB")

    def test_multivalue_field_matching(self):
        """
        Test that field specifications only show those results with the desired
        field values - even when there is an array of possible values
        """
        DemoCourse.get_and_index(self.searcher, {"modes": ["honor", "verified"]})
        DemoCourse.get_and_index(self.searcher, {"modes": ["honor"]})

        results = course_discovery_search()
        self.assertEqual(results["total"], 2)

        results = course_discovery_search(field_dictionary={"modes": "honor"})
        self.assertEqual(results["total"], 2)

        results = course_discovery_search(field_dictionary={"modes": "verified"})
        self.assertEqual(results["total"], 1)

    def test_aggregating(self):
        """
        Tests bucket aggregation results.

        Test that aggregation results are incorporated (org and modes are
        default aggregations).
        """
        DemoCourse.get_and_index(self.searcher, {"org": "OrgA", "modes": ["honor", "verified"]})
        DemoCourse.get_and_index(self.searcher, {"org": "OrgA", "modes": ["honor"]})
        DemoCourse.get_and_index(self.searcher, {"org": "OrgB", "modes": ["honor"]})
        DemoCourse.get_and_index(self.searcher, {"org": "OrgB", "modes": ["verified"]})
        DemoCourse.get_and_index(self.searcher, {"modes": ["other"]})

        results = course_discovery_search()
        self.assertEqual(results["total"], 5)
        self.assertIn("aggs", results)

        self.assertIn("org", results["aggs"])
        self.assertEqual(results["aggs"]["org"]["total"], 4)
        self.assertEqual(results["aggs"]["org"]["terms"]["OrgA"], 2)
        self.assertEqual(results["aggs"]["org"]["terms"]["OrgB"], 2)

        self.assertIn("modes", results["aggs"])
        self.assertEqual(results["aggs"]["modes"]["total"], 6)
        self.assertEqual(results["aggs"]["modes"]["terms"]["honor"], 3)
        self.assertEqual(results["aggs"]["modes"]["terms"]["verified"], 2)
        self.assertEqual(results["aggs"]["modes"]["terms"]["other"], 1)

    @override_settings(COURSE_DISCOVERY_FILTERS=["test_name", "modes"])
    def test_aggregating_filters(self):
        """
        Test that aggregation under consideration can be specified
        by virtue of filters being overriden
        """
        DemoCourse.get_and_index(self.searcher, {"test_name": "Test Name 1", "modes": ["honor", "verified"]})
        DemoCourse.get_and_index(self.searcher, {"test_name": "Test Name 1", "modes": ["honor"]})
        DemoCourse.get_and_index(self.searcher, {"test_name": "Test Name 2", "modes": ["honor"]})
        DemoCourse.get_and_index(self.searcher, {"test_name": "Test Name 2", "modes": ["verified"]})
        DemoCourse.get_and_index(self.searcher, {"modes": ["other"]})

        results = course_discovery_search()
        self.assertEqual(results["total"], 5)
        self.assertIn("aggs", results)

        self.assertNotIn("org", results["aggs"])

        self.assertIn("modes", results["aggs"])
        self.assertEqual(results["aggs"]["modes"]["total"], 6)
        self.assertEqual(results["aggs"]["modes"]["terms"]["honor"], 3)
        self.assertEqual(results["aggs"]["modes"]["terms"]["verified"], 2)
        self.assertEqual(results["aggs"]["modes"]["terms"]["other"], 1)

        self.assertIn("test_name", results["aggs"])
        self.assertEqual(results["aggs"]["test_name"]["total"], 4)
        self.assertEqual(results["aggs"]["test_name"]["terms"]["Test Name 1"], 2)
        self.assertEqual(results["aggs"]["test_name"]["terms"]["Test Name 2"], 2)

    @override_settings(COURSE_DISCOVERY_AGGREGATIONS={"subject": {}, "lang": {}})
    def test_aggregating_override(self):
        """
        Test that aggregation under consideration can be specified
        with custom setting
        """
        DemoCourse.get_and_index(self.searcher, {"subject": "Mathematics", "lang": ["en", "fr"]})
        DemoCourse.get_and_index(self.searcher, {"subject": "Mathematics", "lang": ["en"]})
        DemoCourse.get_and_index(self.searcher, {"subject": "History", "lang": ["en"]})
        DemoCourse.get_and_index(self.searcher, {"subject": "History", "lang": ["fr"]})
        DemoCourse.get_and_index(self.searcher, {"lang": ["de"]})

        results = course_discovery_search()
        self.assertEqual(results["total"], 5)
        self.assertIn("aggs", results)

        self.assertNotIn("org", results["aggs"])
        self.assertNotIn("modes", results["aggs"])

        self.assertIn("subject", results["aggs"])
        self.assertEqual(results["aggs"]["subject"]["total"], 4)
        self.assertEqual(results["aggs"]["subject"]["terms"]["Mathematics"], 2)
        self.assertEqual(results["aggs"]["subject"]["terms"]["History"], 2)

        self.assertIn("lang", results["aggs"])
        self.assertEqual(results["aggs"]["lang"]["total"], 6)
        self.assertEqual(results["aggs"]["lang"]["terms"]["en"], 3)
        self.assertEqual(results["aggs"]["lang"]["terms"]["fr"], 2)
        self.assertEqual(results["aggs"]["lang"]["terms"]["de"], 1)


@override_settings(SEARCH_ENGINE="search.tests.utils.ForceRefreshElasticSearchEngine")
@ddt.ddt
class TestElasticCourseDiscoverySearch(TestMockCourseDiscoverySearch):
    """ version of tests that use Elastic Backed index instead of mock """

    def test_course_matching_empty_index(self):
        """ Check for empty result count before indexing """
        results = course_discovery_search("defensive")
        self.assertEqual(results["total"], 0)

    @ddt.data(
        ('defensive', 1),
        ('offensive', 2),
        ('move', 3),
        ('"offensive move"', 1),
        ('"offensive move"', 1),
        ('"highly-offensive"', 1),
        ('"highly-offensive teams"', 1),
    )
    @ddt.unpack
    # pylint: disable=arguments-differ
    def test_course_matching(self, term, result_count):
        """ Make sure that matches within content can be located and processed """
        DemoCourse.get_and_index(self.searcher, {
            "content": {
                "short_description": "This is a defensive move",
                "overview": "Defensive teams often win"
            }
        })

        DemoCourse.get_and_index(self.searcher, {
            "content": {
                "short_description": "This is an offensive move",
                "overview": "Offensive teams often win"
            }
        })

        DemoCourse.get_and_index(self.searcher, {
            "content": {
                "short_description": "This is a hyphenated move",
                "overview": "Highly-offensive teams often win"
            }
        })

        results = course_discovery_search()

        self.assertEqual(results["total"], 3)

        results = course_discovery_search(term)
        self.assertEqual(results["total"], result_count)


@override_settings(SEARCH_ENGINE=None)
class TestNone(TestCase):
    """ Tests correct skipping of operation when no search engine is defined """

    def test_perform_search(self):
        """ search opertaion should yeild an exception with no search engine """
        with self.assertRaises(NoSearchEngineError):
            course_discovery_search("abc test")



""" Tests for timing functionality """

import datetime
from unittest.mock import patch, call

from django.test import TestCase
from django.test.utils import override_settings
from search.tests.mock_search_engine import MockSearchEngine
from search.utils import Timer
from search.api import emit_api_timing_event


@override_settings(SEARCH_ENGINE="search.tests.mock_search_engine.MockSearchEngine")
class TimingEventsTest(TestCase):
    """ Tests to see if timing events are emitted"""

    def setUp(self):
        super().setUp()
        MockSearchEngine.destroy()
        patcher = patch('search.api.track')
        self.mock_track = patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        MockSearchEngine.destroy()
        super().tearDown()

    def test_perform_search(self):
        search_term = "testing search"
        course_id = "mock.course.id"

        filter_generation_timer = Timer()
        filter_generation_timer.start()
        filter_generation_timer.stop()

        search_timer = Timer()
        search_timer.start()
        search_timer.stop()

        processing_timer = Timer()
        processing_timer.start()
        processing_timer.stop()

        emit_api_timing_event(search_term, course_id, filter_generation_timer, processing_timer, search_timer)
        timing_event_call = self.mock_track.emit.mock_calls[0]
        expected_call = call("edx.course.search.executed", {
            "search_term": search_term,
            "course_id": course_id,
            "filter_generation_time": {
                "start": filter_generation_timer.start_time,
                "end": filter_generation_timer.end_time,
                "elapsed": filter_generation_timer.elapsed_time,
            },
            "processing_time": {
                "start": processing_timer.start_time,
                "end": processing_timer.end_time,
                "elapsed": processing_timer.elapsed_time,
            },
            "search_time": {
                "start": search_timer.start_time,
                "end": search_timer.end_time,
                "elapsed": search_timer.elapsed_time,
            },
        })
        self.assertEqual(timing_event_call, expected_call)


class TimerTest(TestCase):
    """
    Timer Test Case
    """

    def test_start_timer(self):
        timer = Timer()
        timer.start()
        timer.stop()
        self.assertIsNotNone(timer.start_time)
        self.assertIsNotNone(timer.end_time)

    def test_elapsed_time(self):
        # pylint: disable=protected-access

        start = datetime.datetime(2024, 1, 1, 0, 0, 0, 0)
        end = start + datetime.timedelta(seconds=5)

        timer = Timer()
        timer._start_time = start
        timer._end_time = end

        self.assertEqual(timer.elapsed_time, 5)
        self.assertEqual(timer.start_time, start)
        self.assertEqual(timer.end_time, end)

    def test_elapsed_time_string(self):
        # pylint: disable=protected-access

        start = datetime.datetime(2024, 1, 1, 0, 0, 0, 0)
        end = start + datetime.timedelta(seconds=5)

        timer = Timer()
        timer._start_time = start
        timer._end_time = end

        self.assertEqual(timer.elapsed_time, 5)
        self.assertEqual(timer.start_time_string, "2024-01-01T00:00:00")
        self.assertEqual(timer.end_time_string, "2024-01-01T00:00:05")
        self.assertGreaterEqual(timer.end_time, timer.start_time)



"""
This is a search engine for Meilisearch. It implements the edx-search's SearchEngine
API, such that it can be setup as a drop-in replacement for the ElasticSearchEngine. To
switch to this engine, you should run a Meilisearch instance and define the following
setting:

    SEARCH_ENGINE = "search.meilisearch.MeilisearchEngine"

You will then need to create the new indices by running:

    ./manage.py lms shell -c "import search.meilisearch; search.meilisearch.create_indexes()"

For more information about the Meilisearch API in Python, check
https://github.com/meilisearch/meilisearch-python

When implementing a new index, you might discover that you need to list explicit filterable
fields. Typically, you try to index new documents, and Meilisearch fails with the
following response:

    meilisearch.errors.MeilisearchApiError: MeilisearchApiError. Error code: invalid_search_filter.
    Error message: Attribute `field3` is not filterable. Available filterable attributes are:
    `field1 field2 _pk`.

In such cases, the filterable field should be added to INDEX_FILTERABLES below. And you should
then run the `create_indexes()` function again, as indicated above.

This search engine was tested for the following indexes:

1. course_info ("course discovery"):
    - Enable the course discovery feature: FEATURES["ENABLE_COURSE_DISCOVERY"] = True
    - A search bar appears in the LMS landing page.
    - Content is automatically indexed every time a course's "schedule & details" are
      edited in the studio, course content is edited or the "reindex" button is clicked.

2. courseware_content ("courseware search"):
    - Enable the courseware search waffle flag:

        ./manage.py lms waffle_flag --create --everyone courseware.mfe_courseware_search

    - Enable the following feature flags:

        FEATURES["ENABLE_COURSEWARE_INDEX"] = True
        FEATURES["ENABLE_COURSEWARE_SEARCH"] = True

    - Courseware content will be indexed by editing course sections and units.
    - Alternatively, click the "Reindex" button in the Studio.
    - Alternatively, index all courses by running: ./manage.py cms reindex_course --active
    - In the learning MFE, a course search bar appears when opening a course.

Note that the index names could be tuned with the COURSEWARE_INFO_INDEX_NAME and
COURSEWARE_CONTENT_INDEX_NAME settings. However, if you decide to change these settings,
beware that many other applications do not respect them...

When facing issues with Meilisearch during indexing, you may want to look at the
Meilisearch logs. You might notice that some indexing tasks failed. In such cases, you
can troubleshoot these tasks by printing them with:

    ./manage.py lms shell -c "import search.meilisearch; search.meilisearch.print_failed_meilisearch_tasks()"
"""

from copy import deepcopy
from datetime import datetime
import hashlib
import json
import logging
import typing as t

import meilisearch

from django.conf import settings
from django.utils import timezone

from search.search_engine_base import SearchEngine
from search.utils import ValueRange


MEILISEARCH_API_KEY = getattr(settings, "MEILISEARCH_API_KEY", "")
MEILISEARCH_URL = getattr(settings, "MEILISEARCH_URL", "http://meilisearch")
MEILISEARCH_INDEX_PREFIX = getattr(settings, "MEILISEARCH_INDEX_PREFIX", "")


logger = logging.getLogger(__name__)


PRIMARY_KEY_FIELD_NAME = "_pk"
UTC_OFFSET_SUFFIX = "__utcoffset"


# In Meilisearch, we need to explicitly list fields for which we expect to define
# filters and aggregation functions.
# This is different than Elasticsearch where we can aggregate results over any field.
# Here, we list facet fields per index.
# Reference: https://www.meilisearch.com/docs/learn/filtering_and_sorting/search_with_facet_filters
# Note that index names are hard-coded here, because they are hardcoded anyway across all of edx-search.
INDEX_FILTERABLES: dict[str, list[str]] = {
    getattr(settings, "COURSEWARE_INFO_INDEX_NAME", "course_info"): [
        "language",  # aggregate by language, mode, org
        "modes",
        "org",
        "catalog_visibility",  # exclude visibility="none"
        "enrollment_end",  # include only enrollable courses
    ],
    getattr(settings, "COURSEWARE_CONTENT_INDEX_NAME", "courseware_content"): [
        PRIMARY_KEY_FIELD_NAME,  # exclude some specific documents based on ID
        "course",  # search courseware content by course
        "org",  # used during indexing
        "start_date",  # limit search to started courses
    ],
}


class MeilisearchEngine(SearchEngine):
    """
    Meilisearch-compatible search engine. We work very hard to produce an output that is
    compliant with edx-search's ElasticSearchEngine.
    """

    def __init__(self, index=None) -> None:
        super().__init__(index=index)
        self._meilisearch_index: t.Optional[meilisearch.index.Index] = None

    @property
    def meilisearch_index(self) -> meilisearch.index.Index:
        """
        Lazy load meilisearch index.
        """
        if self._meilisearch_index is None:
            meilisearch_index_name = get_meilisearch_index_name(self.index_name)
            meilisearch_client = get_meilisearch_client()
            self._meilisearch_index = meilisearch_client.index(meilisearch_index_name)
        return self._meilisearch_index

    @property
    def meilisearch_index_name(self):
        """
        The index UID is its name.
        """
        return self.meilisearch_index.uid

    def index(self, sources: list[dict[str, t.Any]], **kwargs):
        """
        Index a number of documents, which can have just any type.
        """
        logger.info(
            "Index request: index=%s sources=%s kwargs=%s",
            self.meilisearch_index_name,
            sources,
            kwargs,
        )
        processed_documents = [process_document(source) for source in sources]
        self.meilisearch_index.add_documents(
            processed_documents, serializer=DocumentEncoder
        )

    def search(
        self,
        query_string=None,
        field_dictionary=None,
        filter_dictionary=None,
        exclude_dictionary=None,
        aggregation_terms=None,
        # exclude_ids=None, # deprecated
        # use_field_match=False, # deprecated
        log_search_params=False,
        **kwargs,
    ):  # pylint: disable=too-many-arguments
        """
        See meilisearch docs: https://www.meilisearch.com/docs/reference/api/search
        """
        opt_params = get_search_params(
            field_dictionary=field_dictionary,
            filter_dictionary=filter_dictionary,
            exclude_dictionary=exclude_dictionary,
            aggregation_terms=aggregation_terms,
            **kwargs,
        )
        if log_search_params:
            logger.info("Search query: opt_params=%s", opt_params)
        meilisearch_results = self.meilisearch_index.search(query_string, opt_params)
        processed_results = process_results(meilisearch_results, self.index_name)
        return processed_results

    def remove(self, doc_ids, **kwargs):
        """
        Removing documents from the index is as simple as deleting the the documents
        with the corresponding primary key.
        """
        logger.info(
            "Remove request: index=%s, doc_ids=%s kwargs=%s",
            self.meilisearch_index_name,
            doc_ids,
            kwargs,
        )
        doc_pks = [id2pk(doc_id) for doc_id in doc_ids]
        if doc_pks:
            self.meilisearch_index.delete_documents(doc_pks)


class DocumentEncoder(json.JSONEncoder):
    """
    Custom encoder, useful in particular to encode datetime fields.
    Ref: https://github.com/meilisearch/meilisearch-python?tab=readme-ov-file#custom-serializer-for-documents-
    """

    def default(self, o):
        if isinstance(o, datetime):
            return str(o)
        return super().default(o)


def print_failed_meilisearch_tasks(count: int = 10):
    """
    Useful function for troubleshooting.

    Since indexing tasks are asynchronous, sometimes they fail and it's tricky to figure
    out why. This will print failed tasks to stdout.
    """
    client = get_meilisearch_client()
    for result in client.task_handler.get_tasks(
        {"statuses": "failed", "limit": count}
    ).results:
        print(result)


def create_indexes(index_filterables: t.Optional[dict[str, list[str]]] = None):
    """
    This is an initialization function that creates indexes and makes sure that they
    support the right facetting.

    The `index_filterables` will default to `INDEX_FILTERABLES` if undefined. Developers
    can use this function to configure their own indices.
    """
    if index_filterables is None:
        index_filterables = INDEX_FILTERABLES

    client = get_meilisearch_client()
    for index_name, filterables in index_filterables.items():
        meilisearch_index_name = get_meilisearch_index_name(index_name)
        index = get_or_create_meilisearch_index(client, meilisearch_index_name)
        update_index_filterables(client, index, filterables)


def get_or_create_meilisearch_index(
    client: meilisearch.Client, index_name: str
) -> meilisearch.index.Index:
    """
    Get an index. If it does not exist, create it.

    This will fail with a RuntimeError if we fail to create the index. It will fail with
    a MeilisearchApiError in other failure cases.
    """
    try:
        return client.get_index(index_name)
    except meilisearch.errors.MeilisearchApiError as e:
        if e.code != "index_not_found":
            raise
        task_info = client.create_index(
            index_name, {"primaryKey": PRIMARY_KEY_FIELD_NAME}
        )
        wait_for_task_to_succeed(client, task_info)
        # Get the index again
        return client.get_index(index_name)


def update_index_filterables(
    client: meilisearch.Client, index: meilisearch.index.Index, filterables: list[str]
) -> None:
    """
    Make sure that the filterable fields of an index include the given list of fields.

    If existing fields are present, they are preserved.
    """
    if not filterables:
        return
    existing_filterables = set(index.get_filterable_attributes())
    if set(filterables).issubset(existing_filterables):
        # all filterables fields are already present
        return
    all_filterables = list(existing_filterables.union(filterables))
    task_info = index.update_filterable_attributes(all_filterables)
    wait_for_task_to_succeed(client, task_info)


def wait_for_task_to_succeed(
    client: meilisearch.Client,
    task_info: meilisearch.task.TaskInfo,
    timeout_in_ms: int = 5000,
) -> None:
    """
    Wait for a Meilisearch task to succeed. If it does not, raise RuntimeError.
    """
    task = client.wait_for_task(task_info.task_uid, timeout_in_ms=timeout_in_ms)
    if task.status != "succeeded":
        raise RuntimeError(f"Failed meilisearch task: {task}")


def get_meilisearch_client():
    """
    Return a Meilisearch client with the right settings.
    """
    return meilisearch.Client(MEILISEARCH_URL, api_key=MEILISEARCH_API_KEY)


def get_meilisearch_index_name(index_name: str) -> str:
    """
    Return the index name in Meilisearch associated to a hard-coded index name.

    This is useful for multi-tenant Meilisearch: just define a different prefix for
    every tenant.

    Usually, meilisearch API keys are allowed to access only certain index prefixes.
    Make sure that your API key matches the prefix.
    """
    return MEILISEARCH_INDEX_PREFIX + index_name


def process_document(doc: dict[str, t.Any]) -> dict[str, t.Any]:
    """
    Process document before indexing.

    We make a copy to avoid modifying the source document.
    """
    processed = process_nested_document(doc)

    # Add primary key field
    processed[PRIMARY_KEY_FIELD_NAME] = id2pk(doc["id"])

    return processed


def process_nested_document(doc: dict[str, t.Any]) -> dict[str, t.Any]:
    """
    Process nested dict inside top-level Meilisearch document.
    """
    processed = {}
    for key, value in doc.items():
        if isinstance(value, timezone.datetime):
            # Convert datetime objects to timestamp, and store the timezone in a
            # separate field with a suffix given by UTC_OFFSET_SUFFIX.
            utcoffset = None
            if value.tzinfo:
                utcoffset = value.utcoffset().seconds
            processed[key] = value.timestamp()
            processed[f"{key}{UTC_OFFSET_SUFFIX}"] = utcoffset
        elif isinstance(value, dict):
            processed[key] = process_nested_document(value)
        else:
            # Pray that there are not datetime objects inside lists.
            # If there are, they will be converted to str by the DocumentEncoder.
            processed[key] = value
    return processed


def id2pk(value: str) -> str:
    """
    Convert a document "id" field into a primary key that is compatible with Meilisearch.

    This step is necessary because the "id" is typically a course id, which includes
    colon ":" characters, which are not supported by Meilisearch. Source:
    https://www.meilisearch.com/docs/learn/getting_started/primary_key#formatting-the-document-id
    """
    return hashlib.sha1(value.encode()).hexdigest()


def get_search_params(
    field_dictionary=None,
    filter_dictionary=None,
    exclude_dictionary=None,
    aggregation_terms=None,
    **kwargs,
) -> dict[str, t.Any]:
    """
    Return a dictionary of parameters that should be passed to the Meilisearch client
    `.search()` method.
    """
    params: dict[str, t.Any] = {"showRankingScore": True}

    # Aggregation
    if aggregation_terms:
        params["facets"] = list(aggregation_terms.keys())

    # Exclusion and inclusion filters
    filters = []
    if field_dictionary:
        filters += get_filter_rules(field_dictionary)
    if filter_dictionary:
        filters += get_filter_rules(filter_dictionary, optional=True)
    if exclude_dictionary:
        filters += get_filter_rules(exclude_dictionary, exclude=True)
    if filters:
        params["filter"] = filters

    # Offset/Size
    if "from_" in kwargs:
        params["offset"] = kwargs["from_"]
    if "size" in kwargs:
        params["limit"] = kwargs["size"]

    return params


def get_filter_rules(
    rule_dict: dict[str, t.Any], exclude: bool = False, optional: bool = False
) -> list[str]:
    """
    Convert inclusion/exclusion rules.
    """
    rules = []
    for key, value in rule_dict.items():
        if isinstance(value, list):
            for v in value:
                rules.append(
                    get_filter_rule(key, v, exclude=exclude, optional=optional)
                )
        else:
            rules.append(
                get_filter_rule(key, value, exclude=exclude, optional=optional)
            )
    return rules


def get_filter_rule(
    key: str, value: str, exclude: bool = False, optional: bool = False
) -> str:
    """
    Meilisearch filter rule.

    See: https://www.meilisearch.com/docs/learn/filtering_and_sorting/filter_expression_reference
    """
    prefix = "NOT " if exclude else ""
    if key == "id":
        key = PRIMARY_KEY_FIELD_NAME
        value = id2pk(value)
    if isinstance(value, str):
        rule = f'{prefix}{key} = "{value}"'
    elif isinstance(value, ValueRange):
        constraints = []
        lower = value.lower
        if isinstance(lower, timezone.datetime):
            lower = lower.timestamp()
        upper = value.upper
        if isinstance(upper, timezone.datetime):
            upper = upper.timestamp()
        # I know that the following fails if value == 0, but we are being
        # consistent with the behaviour in the elasticsearch engine.
        if lower:
            constraints.append(f"{key} >= {lower}")
        if upper:
            constraints.append(f"{key} <= {upper}")
        rule = " AND ".join(constraints)
        if len(constraints) > 1:
            rule = f"({rule})"
    else:
        raise ValueError(f"Unknown value type: {value.__class__}")
    if optional:
        rule += f" OR {key} NOT EXISTS"
    return rule


def process_results(results: dict[str, t.Any], index_name: str) -> dict[str, t.Any]:
    """
    Convert results produced by Meilisearch into results that are compatible with the
    edx-search engine API.

    Example input:

        {
            'hits': [
                {
                    'pk': 'f381d4f1914235c9532576c0861d09b484ade634',
                    'id': 'course-v1:OpenedX+DemoX+DemoCourse',
                    ...
                    "_rankingScore": 0.865,
                },
                ...
            ],
            'query': 'demo',
            'processingTimeMs': 0,
            'limit': 20,
            'offset': 0,
            'estimatedTotalHits': 1
        }

    Example output:

        {
                'took': 13,
                'total': 1,
                'max_score': 0.4001565,
                'results': [
                    {
                        '_index': 'course_info',
                        '_type': '_doc',
                        '_id': 'course-v1:OpenedX+DemoX+DemoCourse',
                        '_ignored': ['content.overview.keyword'], # removed
                        'data': {
                            'id': 'course-v1:OpenedX+DemoX+DemoCourse',
                            'course': 'course-v1:OpenedX+DemoX+DemoCourse',
                            'content': {
                                'display_name': 'Open edX Demo Course',
                                ...
                            },
                            'image_url': '/asset-v1:OpenedX+DemoX+DemoCourse+type@asset+block@thumbnail_demox.jpeg',
                            'start': '2020-01-01T00:00:00+00:00',
                            ...
                        },
                        'score': 0.4001565
                    }
                ],
                'aggs': {
                    'modes': {
                        'terms': {'audit': 1},
                        'total': 1.0,
                        'other': 0
                    },
                    'org': {
                        'terms': {'OpenedX': 1}, 'total': 1.0, 'other': 0
                    },
                    'language': {'terms': {'en': 1}, 'total': 1.0, 'other': 0}
                }
            }
    """
    # Base
    processed = {
        "took": results["processingTimeMs"],
        "total": results["estimatedTotalHits"],
        "results": [],
        "aggs": {},
    }

    # Hits
    max_score = 0
    for result in results["hits"]:
        result = process_hit(result)
        score = result.pop("_rankingScore")
        max_score = max(max_score, score)
        processed_result = {
            "_id": result["id"],
            "_index": index_name,
            "_type": "_doc",
            "data": result,
        }
        processed["results"].append(processed_result)
    processed["max_score"] = max_score

    # Aggregates/Facets
    for facet_name, facet_distribution in results.get("facetDistribution", {}).items():
        total = sum(facet_distribution.values())
        processed["aggs"][facet_name] = {
            "terms": facet_distribution,
            "total": total,
            "other": 0,
        }
    return processed


def process_hit(hit: dict[str, t.Any]) -> dict[str, t.Any]:
    """
    Convert a search result back to the ES format.
    """
    processed = deepcopy(hit)

    # Remove primary key field
    try:
        processed.pop(PRIMARY_KEY_FIELD_NAME)
    except KeyError:
        pass

    # Convert datetime fields back to datetime
    for key in list(processed.keys()):
        if key.endswith(UTC_OFFSET_SUFFIX):
            utcoffset = processed.pop(key)
            key = key[: -len(UTC_OFFSET_SUFFIX)]
            timestamp = hit[key]
            tz = (
                timezone.get_fixed_timezone(timezone.timedelta(seconds=utcoffset))
                if utcoffset
                else None
            )
            processed[key] = timezone.datetime.fromtimestamp(timestamp, tz=tz)
    return processed



""" import urls from search component to test it's operation when included within other django projects """

import django
from django.urls import include, path

# from django.contrib import admin
# admin.autodiscover()

import search

# urlpatterns is the standard name to use here
# pylint: disable=invalid-name
urlpatterns = [path('search/', include(search.urls))]



#!/usr/bin/env python
""" Setup to allow pip installs of edx-search module """
import os
import re
from pathlib import Path

from setuptools import setup


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns:
        list: Requirements file relative path strings
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.split('#')[0].strip()
            for line in open(path, encoding='utf-8').readlines()  # pylint: disable=consider-using-with
            if is_requirement(line.strip())
        )
    return list(requirements)


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.
    Returns:
        bool: True if the line is not blank, a comment, a URL, or an included file
    """
    return line and not line.startswith(('-r', '#', '-e', 'git+', '-c'))


def get_version(*file_paths):
    """
    Extract the version string from the file at the given relative path fragments.
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename, encoding='utf-8') as opened_file:
        version_file = opened_file.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


VERSION = get_version('edxsearch', '__init__.py')

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='edx-search',
    version=VERSION,
    description='Search and index routines for index access',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='edX',
    author_email='oscm@edx.org',
    url='https://github.com/openedx/edx-search',
    license='AGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.2',
    ],
    packages=['search', 'search.tests', 'edxsearch'],
    install_requires=load_requirements('requirements/base.in')
)
